package se.umu.jobsubmission.util;

import java.util.List; 
import java.util.LinkedList; 
import se.umu.wsag.client.ReservationClient; 
import org.apache.axis.message.addressing.EndpointReferenceType;

import se.umu.reservation.ReservationType;
import se.umu.reservation.ReservationInfo; 

import org.apache.log4j.Logger;

/**
 * @author tordsson
 * 
 * Contains utils for handling created reservation from a broker
 * perspective. All created reservations are stored within this class, 
 * as
 *
 */
public class ReservationUtils {
	
	private static Logger logger = Logger.getLogger(ReservationUtils.class); 
	
	ReservationClient reservationClient;  
	List reservations; 
	
	public ReservationUtils() {
		this.reservationClient = new ReservationClient(); 	
		this.reservations = new LinkedList(); 	
	}		
    
    /**
     * Confirms reservation (commits it), by extending lifetime 
     * to ensure that reservation resource will exist until reservation 
     * expires. 
     * @param object The resource where the reservation is created. 
     */
    public void confirmReservation(Object resource) throws Exception {
    	
    	ReservationInfo info = getReservationInfo(resource); 
    	if(info == null) {
    		logger.debug("No reservation found for " + resource); 
    		throw new Exception("No reservation found"); 
    	}
    	EndpointReferenceType epr = info.getEpr();     
    	this.reservationClient.commitReservation(epr); 
    	
    	logger.debug("Committed reservation: " 
    			+ info.getReservation().getLocalID()); 
    	
    	// Mark reservation as committed, to ensure that 
    	// it is not removed from the repository. 
    	
    	// This is kind of cheating, we should actually be retrieving a 
    	// new copy of the resource properties...
    	info.getReservation().setCommitted(true); 
    }
	
    /**
     * Store reservation for later use. 
     * @param epr
     * @param resource
     * @param reservationTerms
     */
	public void storeReservation(EndpointReferenceType epr, 
			Object resource, 
			ReservationType reservationTerms) {
		
		logger.debug("Storing reservation " + epr + "\n with key: " + resource); 
		
		ReservationInfo resInfo = new ReservationInfo(); 
		resInfo.setEpr(epr); 
		resInfo.setResource(resource); 
		resInfo.setReservation(reservationTerms); 
		 
		reservations.add(resInfo); 
		
	}
	
    /**
     * Removes a reservation that will not be used. 
     * @param resource The resource where the reservation to be removed
     * is created. Also removes the reservation from the repository     
     * and destroys the WS-Resource representing the reservation. 
     * @throws Exception
     */
	public void removeReservation(Object resource) {
		
		ReservationInfo info = this.getReservationInfo(resource); 
		if(info == null) {
			logger.warn("Error removing reservation for resource: " + resource); 
			return; 
		}
		String localID = info.getReservation().getLocalID(); 
		logger.debug("Removing reservation: " + localID); 
		try {
			reservationClient.destroyReservation(info.getEpr()); 
		} catch (Exception e) {
			logger.warn("Error removing reservation: " + localID); 
		}
		this.reservations.remove(info); 
		
		logger.debug("We have " + this.reservations.size() 
				+ " reservations left"); 
	}
	
	/**
	 * @param epr key to search for. 
	 * @return ReservationInfo associated with the given epr, or null
	 * if no reservation is found. 
	 */
	public ReservationInfo getReservationInfo(EndpointReferenceType epr) {
		
		ReservationInfo tmp; 
		for(int i = 0; i < reservations.size(); i++) {
			tmp = (ReservationInfo) reservations.get(i); 
			if(tmp.getEpr().equals(epr))
				return tmp; 
		}
		
		return null; 
	}
	
	/**
	 * 
	 * @param resource The key to search for. 
	 * @return Returns all reservations created at the resource passed 
	 * as argument. 
	 */
	public ReservationInfo getReservationInfo(Object resource) {
		
		ReservationInfo tmp; 
		this.printRes(reservations); 
		for(int i = 0; i < reservations.size(); i++) {
			tmp = (ReservationInfo) reservations.get(i); 
			if(tmp.getResource().equals(resource))  
				return tmp; 
		}
		
		return null; 
	}
	
	public List getAllReservations() {
		return this.reservations; 
	}
	


    
    public void printRes(List reservations) {
    
    	ReservationInfo reservationInfo; 
    
    	logger.debug("We got " + reservations.size() + " reservations:"); 
    	for(int i = 0; i < reservations.size(); i++)
    		logger.debug(((ReservationInfo)reservations.get(i)).
    				getReservation().getLocalID()); 
    }
    
    /**
     * Clears the reservation repository (typically, done after 
     * job submission is successful). However, keep the committed reservations, 
     * as these are used by submitted jobs...
     * @param resource Save the reservation for this one. 
     */
    public void removeUnusedReservations() { // Object selectedResource) {
    	
    	logger.debug("Removing unused reservations");     	
    	
    	printRes(reservations); 
    	ReservationInfo reservationInfo; 
    	
    	for(int i = 0; i < reservations.size(); i++) {
    		
    		reservationInfo = (ReservationInfo) reservations.get(i); 
    		boolean committed  = reservationInfo.getReservation().isCommitted(); 
    		String localID =  reservationInfo.getReservation().getLocalID();
    		if(committed)
    			logger.debug("Keeping reservation: " + localID); 
    		else {
    			logger.debug("Removing reservation: " + localID); 
    			removeReservation(reservationInfo.getResource()); 
    		}
    	}
    }
	
			
	
}	
		
