/*
 * Created on Feb 24, 2006
 *
 */
package se.umu.jobsubmission.impl.submitter;

import java.io.FileInputStream; 
import java.io.StringWriter; 
import java.util.Arrays;
import java.util.Calendar; 
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map; 
import java.util.HashMap; 

import org.globus.wsrf.config.ContainerConfig;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobIdentification_Type;
import org.globus.mds.glue.GLUECERPType;

import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.utils.FaultHelper;

import org.xml.sax.InputSource;

import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.SubmittedJob;
import se.umu.jobsubmission.impl.broker.Broker;
import se.umu.jobsubmission.impl.broker.RunTimePredictor;
import se.umu.jobsubmission.impl.broker.Selector;
import se.umu.jobsubmission.impl.dispatcher.Dispatcher;
import se.umu.jobsubmission.impl.informationfinder.InformationFinder;
import se.umu.jobsubmission.impl.informationfinder.InformationFinderFactory;
import se.umu.jobsubmission.util.ReservationUtils;
import se.umu.jobsubmission.util.SubmitterUtils;
import se.umu.reservation.ReservationType;
import se.umu.reservation.ReservationInfo; 
import se.umu.wsag.client.ReservationClient;
import se.umu.wsag.impl.AgreementFactoryConstants;
import se.umu.wsag.util.JobSubmissionUtils;

import org.ggf.ws_agreement.ContinuingFaultType;

import se.umu.jobsubmission.util.JSDLUtils; 

import se.umu.jobsubmission.util.ClientUtils;

import se.umu.jobsubmission.ClientConfiguration;
import org.globus.wsrf.encoding.ObjectDeserializer;


/**
 * @author tordsson
 *
 * An alternative submitter that performs coallocation. 
 * Given a input of a set of jobs (at least 2), 
 * resources are reserved for each job. Each reservation 
 * is created witihn a given interval of time and 
 * no reservation deviates from another with more than a 
 * specified tolerance level. 
 * 
 * Next, each subjob is submitted ...
 *
 */
public class Coallocator {
	
	// TODO: These should preferably be configurable... 
	public static final int DEFAULT_TOLERANCE = 60; 
	public static final int DEFAULT_WINDOW_SIZE = 3600; 
		
	// Only used to test path augmentation
	boolean mayCreateNewAtToggle02; 
	
	/** 
	 * Class used to sort job input entries, placing the most demanding 
	 * job first. 
	 */
	private class RunTimeComparator implements Comparator {
		
		/* 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 * Compares run times of jobs 
		 */
		public int compare(Object o1, Object o2) throws ClassCastException {
			
			CoallocationInputEntry entry1 = (CoallocationInputEntry) o1; 
	    	CoallocationInputEntry entry2 = (CoallocationInputEntry) o2; 
	    	
	    	// FIXME: Which runtime to use. The run-time may differ 
	    	// for different resources in the list of rankedResources... 
	    	// Take worst-case? 
	    	// 	-> Use mean value of all runtimes...
	    	
	    	int time1 = JSDLUtils.getDuration(entry1.getJobDescription()); 
	    	int time2 = JSDLUtils.getDuration(entry2.getJobDescription()); 
	    	
	    	// N.B. We want largest run time first, that is, 
	    	// reversed sorting order. 
	    	// TODO: This is the correct code
			return time2 - time1; 
	    	
	    	// Sort them backwards just for testing...
	    	// TODO: Remove this incorrect code later
	    	// return time1 - time2; 
	    	
		}
	}
	
	// Compares two jobs, one is smaller if it has more CPUs... 
	// This ensures that sorting is done largest job first... 
	private class NumberOfCPUsComparator implements Comparator {
		
		public int compare(Object o1, Object o2) throws ClassCastException {
			CoallocationInputEntry entry1 = (CoallocationInputEntry) o1; 
	    	CoallocationInputEntry entry2 = (CoallocationInputEntry) o2; 
	    	
	    	int numberOfCPUs1 = JSDLUtils.getNumberOfCPUs(entry1.getJobDescription()); 
	    	int numberOfCPUs2 = JSDLUtils.getNumberOfCPUs(entry2.getJobDescription()); 
	    	
	    	return numberOfCPUs2 - numberOfCPUs1; 
		}
	}
	
	// Compares two jobs based on memory requirement, 
	/// more memory means "smaller", ensures largest memory first in sorting... 
	private class MemoryComparator implements Comparator {
		
		public int compare (Object o1, Object o2) throws ClassCastException {
			CoallocationInputEntry entry1 = (CoallocationInputEntry) o1; 
			CoallocationInputEntry entry2 = (CoallocationInputEntry) o2; 
			
			int memory1 = JSDLUtils.getMemory(entry1.getJobDescription()); 
			int memory2 = JSDLUtils.getMemory(entry2.getJobDescription()); 
			
			return memory2 - memory1; 
		}
	}

    private static Logger logger = Logger.getLogger(Coallocator.class); 

    private InformationFinderFactory informationFinderFactory; 
    private String dispatcherClasses[]; 
    private String dataManagerClasses[]; 
    private String stagingTargetResolverClasses[]; 
    private int numberOfThreads;     
    
    /**
     * From JobSubmissionService, first create submitter, then 
     * create coallocator and let it share the config of the submitter... 
     * @param factory
     * @param dispatcherClasses
     * @param dataManagerClasses
     * @param numberOfThreads
     */
    public Coallocator(InformationFinderFactory factory, 
    		String dispatcherClasses[],
			String dataManagerClasses[], 
			int numberOfThreads) throws Exception {
    	this.informationFinderFactory = factory; 
    	this.dispatcherClasses = dispatcherClasses; 
    	this.dataManagerClasses = dataManagerClasses; 
    	this.numberOfThreads = numberOfThreads; 
    	
    	// Get ClientJobStaging implementations (used to cancel jobs)
    	String configFile = "/etc/jobsubmission/clientconfig.xml"; 
    	String globusLocation = ContainerConfig.getGlobusLocation(); 
    	String filePath = globusLocation + configFile; 

        ClientConfiguration config = null;
        
        FileInputStream input = new FileInputStream(filePath); 
        InputSource inputSource = new InputSource(input); 
        config = (ClientConfiguration) ObjectDeserializer.deserialize(inputSource, 
        		ClientConfiguration.class); 
        
        this.stagingTargetResolverClasses = config.getStagingTargetResolver(); 

    	
    }
    
    /**
     * Add all preselected resources to a list. 
     * @param jobPreferences
     * @return
     */
    public String[] getPreselectedResources(JobPreferencesType jobPreferences[]) {
    	
    	if(jobPreferences == null || jobPreferences.length == 0) {
    		return null; 
    	}
    	
        // Check if pre-selected resources targets are specified: 
    	// Check if submission target is defined in any of the 
    	// job preferences files. 
    	String resourceURLs[] = null; 
    	List targets = new LinkedList(); 
    	for(int i = 0; i < jobPreferences.length; i++) {
    		try {
    			resourceURLs = jobPreferences[i].getJobRequirement().
					getSubmissionTarget(); 
    			if(resourceURLs != null) {
    				for(int j = 0; j < resourceURLs.length; j++) {
    					targets.add(resourceURLs[j]); 
    				}
    			}
    		} catch (Exception e) {
    			resourceURLs = null; 
    		}    
    	}
    	if(targets.size() > 0) {
    		String results[] = new String[targets.size()]; 
    		for(int i = 0; i < targets.size(); i++) {
    			results[i] = (String) targets.get(i); 
    		}
    		return results; 
    	}
    	else {
    		return null;  
    	}
    }

    
    /**
     * Performs resource discovery and information retrieval. 
     * For each job description, filter and rank the resources 
     * suitable for the job. Returns an arrays of lists, 
     * where each list contains the resources suitable for the 
     * corresponding job. 
     * 
     * @param jobDescriptions
     * @param indexServerURIs
     * @param jobPreferences
     * @param broker
     * @return
     * @throws Exception
     */
    public List[] rankResources(JobDefinition_Type jobDescriptions[], 
    		String indexServerURIs[], 
			JobPreferencesType jobPreferences[], 
			Broker broker) throws Exception {
    	
    	String resourceURLs[] = getPreselectedResources(jobPreferences); 
		
		// Get resource info 
		InformationFinder informationFinder = informationFinderFactory.
		getInformationFinder(); 
		List foundResources = SubmitterUtils.getResources(informationFinder, 
				indexServerURIs, 
				resourceURLs, 
				broker); 
		
		// Filter resources, this throws an exception if 
		// no resources are suitable. 
		List approvedResources[] = new List[jobDescriptions.length];        
		for(int i = 0; i < jobDescriptions.length; i++) {
			approvedResources[i] = SubmitterUtils.filterResources(foundResources, 
					jobDescriptions[i], 
					broker); 
		}		
				
		// Rank resources 
		List rankedResources[] = new List[jobDescriptions.length]; 
		for(int i = 0; i < jobDescriptions.length; i++) {				
			Selector selector = broker.getSelector(jobPreferences[i]); 			
			rankedResources[i] = broker.select(approvedResources[i], 
					jobDescriptions[i], 
					jobPreferences[i], 
					selector);
		}		 	

		return rankedResources; 
    	
    }
    
    /**
     * Disables reservations in resource ranking by setting "useReservation"
     * to false in the job preferences document. This is done no matter
     * whether the useReservation part of the job preferences document 
     * is specified. 
     * @param pref
     * @return
     */
    public JobPreferencesType disableReservation(JobPreferencesType pref) {
    	
    	JobRequirementType requirements = pref.getJobRequirement();
    	if(requirements == null) {
    		requirements = new JobRequirementType(); 
    		pref.setJobRequirement(requirements); 
    	} 
    	Boolean useReservations = requirements.getUseReservations(); 
    	if(useReservations == null || useReservations.booleanValue()) {
    		useReservations = new Boolean(false); 
    		requirements.setUseReservations(useReservations); 
    	}
    	
    	return pref; 
    }
    
    /**
     * Constructs a dummy job preferences document. Such a document is
     * associated with job descriptions lacking a preferences document, 
     * thus indicating that this job may start concurrently with the 
     * other jobs. 
     * @return
     */
    public JobPreferencesType makeDummyPreferences(Calendar early, Calendar late) {
    	
    	JobPreferencesType preferences = new JobPreferencesType(); 
    	JobRequirementType requirements = new JobRequirementType(); 
    	requirements.setEarliestAbsoluteStart(early); 
    	requirements.setLatestAbsoluteStart(late);    
    	preferences.setJobRequirement(requirements);     	
    	
    	return preferences; 
    }
    
    /**
     * Creates a job preferences document for each job. 
     * For those jobs that lack such a document, a dummy document
     * is created
     * @param jobPreferences
     * @param numberOfJobs
     * @return
     */
    public JobPreferencesType[] fixJobPreferences(
    		JobPreferencesType jobPreferences[], 
    		int numberOfJobs) {
    	
    	// Find earliest (and latest) allowed start 
  		Calendar earliest = null; 
  		Calendar latest = null; 
  		
		if(jobPreferences != null) {
			for(int i = 0; i < jobPreferences.length; i++) {
				try {
					
					// Translates relative start time to absolute start time
					jobPreferences[i] = SubmitterUtils.fixPreferences(jobPreferences[i]); 
					
					Calendar earlyCal = jobPreferences[i].getJobRequirement().getEarliestAbsoluteStart();
					if(earliest == null || earlyCal.before(earliest)) {
						earliest = earlyCal; 
					}
					
					Calendar lateCal = jobPreferences[i].getJobRequirement().getLatestAbsoluteStart(); 
					if(latest == null || lateCal.after(latest)) {
						latest = lateCal; 
					}
										
				} catch (Exception e){
					logger.warn("Ooops: " + e.getMessage()); 
					// TODO: Now what? 
					// -> This is the result of a NPE (some part of job pref is missing) 
				} 
			}
		}		
		
		if(earliest == null) {
			earliest = Calendar.getInstance(); 
		}
		if(latest == null) {
			latest = Calendar.getInstance(); 
			latest.add(Calendar.SECOND, Coallocator.DEFAULT_WINDOW_SIZE); 
		}
		
		// Create dummy documents for those jobs not specifying one. 
		JobPreferencesType result[] = new JobPreferencesType[numberOfJobs]; 
		for(int i = 0; i < result.length; i++) {			
			if(jobPreferences != null && i < jobPreferences.length) {
				logger.warn("Using existing preferences"); 
				result[i] = jobPreferences[i]; 
			}
			else {
				JobPreferencesType noPreferences = makeDummyPreferences(earliest, latest); 
				logger.warn("creating new preferences"); 
				result[i] = noPreferences; 
				
			}
			result[i] = disableReservation(result[i]); 
		}
		
		return result; 
    }
    
    /**
     * Returns true if some other job has reserved a resource the could
     * be used for this job, false otherwise. 
     * @param allJobs
     * @param thisJobs
     * @param existingReservation
     * @return
     * @throws Exception
     */
    public boolean augmentationPossible(CoallocationInputEntry allJobs[], 
    		CoallocationInputEntry thisJob, 
			Map existingReservations) throws Exception {
    	
		List possibleResources = thisJob.getRankedResources(); 
		
		for(int i = 0; i < possibleResources.size(); i++) {			
			GLUECERPType resource = (GLUECERPType) possibleResources.get(i); 				
			List competingJobs = (List) existingReservations.get(resource); 
			if(competingJobs != null) {
				return true;  
			}
		}
    	
    	return false; 
    }
    
    
    /**
     * Finds an augmenting path by performing BFS in the bipartite graph
     * that is formed by the jobs and the resources. 
     * @param allJobs 
     * @param job
     * @param early
     * @param late
     * @param reservationClient
     * @param resourceReservations A map (GLUECERPType -> ReservationInfo) 
     * associating a resource with the
     * reservations currently created for that resource. 
     * @return
     * @throws Exception
     * 
     */
	public List findPath(CoallocationInputEntry allJobs[],
			CoallocationInputEntry thisJob, 
			Calendar early, 
			Calendar late, 
			Map existingReservations) throws Exception {

		List searchQueue = new LinkedList(); 
		
		List firstEntry = new LinkedList(); 
		firstEntry.add(thisJob); 		
		searchQueue.add(firstEntry); 
				
		logger.warn("Augmenting reservation, starting with path: "); 
		printPath(firstEntry); 
		
		// Perform BFS in the bipartite graph, expand path two steps at the time
		// as edges in graph have quite different meaning... 
		while(searchQueue.size() > 0) {
			
			try { 
				// logger.warn("\nNew round"); 
				// Thread.currentThread().sleep(200);
			} catch (Exception e) {} 
			
			List currentEntry = (List) searchQueue.remove(0);  	
			// logger.warn("Current entry is: "); 
			
			// Get last job in current path
			CoallocationInputEntry currentJob = (CoallocationInputEntry) currentEntry.get(
					currentEntry.size() - 1); 		
			
			// All resources this job could use, this list can not be null 
			// at this stage, job would have failed earlier. 
			List possibleResources = currentJob.getRankedResources(); 
								
			// Find out if any other job has reserved a resource that currentjob
			// may be able to use			
			// Expand (job) to (job, R1), (job, R2), ... , (job, RN) and	        
			// For each entry (job, Ri) find 
			// (job, Ri, jobX), (job, Ri, JobY), ... , (job, Ri, JobZ) 
			// where job can use Ri, and jobX has reserved Ri			
			for(int i = 0; i < possibleResources.size(); i++) {				
				
				GLUECERPType resource = (GLUECERPType) possibleResources.get(i); 				
				List competingJobs = (List) existingReservations.get(resource); 
				String resourceName = resource.getCluster(0).getSubCluster(0).getHost(0).getName();				
				
				if(competingJobs == null || competingJobs.size() == 0) {
					
					logger.warn("There are no competing jobs for resource " + resourceName);
					
					// This corresponds to case *)  
					if(currentEntry.size() == 1) {
						logger.warn("Path leads to dead-end, dropping it");
					}					
					// This also corresponds to case ***)
					else {
						logger.warn("Non-trivial augmenting path found"); 
						currentEntry.add(resource); 
						return currentEntry; 
					}					
				}
				else {
					// Now for (job, Ri), form a list 
					// (job, Ri, jobX), (job, Ri, jobY), ..., (job, Ri, jobZ) 
					// where job != jobX, job!=jobY, ..., job!=jobZ 
					// that is, we avoid cycles in the graph... 
					for(int j = 0; j < competingJobs.size(); j++) {
						
						CoallocationInputEntry competingJob = (CoallocationInputEntry) competingJobs.get(j); 
						if(competingJob.equals(currentJob)) {
							logger.warn("Path leads to cycle, dropping it"); 
						}
						else {
							logger.warn("Adding new edge to queue: " + resourceName + 
							 		" , " + competingJob.getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
						
							List newEntry = new LinkedList(); 
							newEntry.addAll(currentEntry); 
							newEntry.add(resource); 
							newEntry.add(competingJob); 							
							 
							// logger.warn("New entry:"); 
							if(validPath(newEntry)) {
								logger.warn("Adding new entry: "); 
								printPath(newEntry); 
								searchQueue.add(newEntry); 
							} 
						}
					}
				}
			}			
		} 
		
		logger.warn("No augmenting path found..."); 
		
		throw new Exception("No augmented path found"); 
			
	}
	
	/**
	 * Checks whether a proposed augmenting path contains loops
	 * @param path
	 * @return
	 */
	public boolean validPath(List path) {
		
		List previous = new LinkedList(); 
		
		for(int i = 0; i < path.size(); i++) {
			Object obj = path.get(i); 
			if(previous.contains(obj)) { 				
				return false; 
			} 
			else { 
				previous.add(obj); 
			}
		}		
	
		return true; 		
		
	}
	
	/**
	 * For debug purposes only
	 * @param path
	 */
	public void printPath(List path) {
		if(path != null) {
			String pathString = ""; 
			for(int i = 0; i < path.size(); i++) {
				Object obj = path.get(i); 
				if(obj instanceof GLUECERPType) {
					pathString += ((GLUECERPType)obj).getCluster(0).getSubCluster(0).getHost(0).getName(); 
				}
				if(obj instanceof CoallocationInputEntry) {
					pathString += ((CoallocationInputEntry) obj).getJobDescription().getJobDescription().getJobIdentification().getJobName(); 
				}
				if(i != path.size() - 1) {
					pathString += " --- "; 
				}
				
			}
			logger.warn(pathString); 
		} 
		else { 
			logger.warn("Path is null!"); 
		}
	}
	
	/**
	 * Apply reservation changes as suggested by the augmenting path, 
	 * thus creating one more reservation. 
	 * @param path List of (job1, reservation2, job2, ... , jobN, reservationN+1) 
	 * @param allJobs
	 * @param resourceReservations
	 * @throws Exception
	 * @throws ContinuingFaultType
	 */
	public void updateReservations(List path, 
			CoallocationInputEntry allJobs[], 
			Calendar early, 
			Calendar late, 
			ReservationClient reservationClient,
			Map resourceReservations) throws Exception, ContinuingFaultType {		
		
		logger.warn("Updating path"); 				
		
		this.mayCreateNewAtToggle02 = true; 
		
		// Path has format
		// <Job 3> --- <Resource 1> -R- <Job 1> --- <Resource 2> -R- <Job 0> --- <Resource 3> 
		// This becomes rather messy, split up path in two lists, jobs and resources. 
		
		List resources = new LinkedList(); 
		List jobs = new LinkedList(); 		
		for(int i = 0; i < path.size();) {
			jobs.add(path.get(i++)); 
			resources.add(path.get(i++)); 
		}
		
		logger.warn("Jobs..."); 
		String jobNames = ""; 
		for(int i = 0; i < jobs.size(); i++) {
			CoallocationInputEntry entry = (CoallocationInputEntry) jobs.get(i); 
			jobNames += "" + entry.getJobDescription().getJobDescription().getJobIdentification().getJobName() + "   "; 		
		}
		logger.warn(jobNames + "\n\n"); 
		
		String resourceNames = ""; 
		for(int i = 0; i < jobs.size(); i++) {
			GLUECERPType jobInfo = (GLUECERPType) resources.get(i); 
			resourceNames += jobInfo.getCluster(0).getSubCluster(0).getHost(0).getName() + "   "; 			
		}
		logger.warn(resourceNames); 
		logger.warn("... and resources"); 					
		
		/* l - last, 
		 * c - current, 
		 * J - job, 
		 * R - resource 
		 * Res - reservation
		 * 
		 * 		Before: 			After: 
		 * 		lJ		cJ			lJ		 cJ
		 * 			   . 			. 		 .
		 *          . cRes 			.old Res . new Res
		 *        . 				.		 .
		 * 		lR		cR			lR		 cR
		 * 
		 * 
		 * 
		 */
						
		int index = resources.size() - 1; 
		GLUECERPType currentResource = (GLUECERPType) resources.get(index); 		
		CoallocationInputEntry currentJob = (CoallocationInputEntry) jobs.get(index); 
		
		// Store old reservation		
		ReservationInfo oldReservation = currentJob.getReservation();
		
		logger.warn("Creating new reservation for job " + currentJob.getJobDescription().
				getJobDescription().getJobIdentification().getJobName()); 
		logger.warn("At resource " + currentResource.getCluster(0).getSubCluster(0).
				getHost(0).getName()); 				
		
		// Create new reservation from cJ to cR 
  		ReservationInfo newReservation = createReservation(currentJob, 
				early, 
				late, 
				reservationClient, 
				currentResource); 
  		currentJob.setReservation(newReservation); 
  		
  		// Add new mapping: 
  		GLUECERPType resource = (GLUECERPType) newReservation.getResource(); 
		List reservedJobs = (List) resourceReservations.get(resource); 
		if(reservedJobs == null) {
			reservedJobs = new LinkedList(); 
		}
		reservedJobs.add(currentJob);     						
		resourceReservations.put(resource, reservedJobs); 
		
		// Store job for later use (removal of mapping) 
		CoallocationInputEntry oldJob = currentJob; 
		
  		// Now that last job has a new reservation, modify all 
  		// existing reservations to fit the previous job instead. 
		do {									
			
			logger.warn("Modifying reservation created for job " + currentJob.getJobDescription().
					getJobDescription().getJobIdentification().getJobName()); 
			
			index--; 
			currentJob = (CoallocationInputEntry) jobs.get(index); 
			currentResource = (GLUECERPType) resources.get(index); 
			
			logger.warn("To suit job " + currentJob.getJobDescription().
					getJobDescription().getJobIdentification().getJobName()); 
			
			logger.warn("Reservation was originally created at resource " 
					+ currentResource.getCluster(0).getSubCluster(0).getHost(0).getName()); 		
			 
			// Create new reservation for the currentjob by modifying the old
			// reservation (the reservation for the previous job). 
			JobDefinition_Type currentJobDefinition = currentJob.getJobDescription(); 
			EndpointReferenceType modifiedReservationEpr = reservationClient.modifyReservation(
					oldReservation.getEpr(), 
					JSDLUtils.getNumberOfCPUs(currentJobDefinition),
					early, 
					late, 
					JSDLUtils.getDuration(currentJobDefinition), 
					JSDLUtils.getFlexible(currentJob.getJobPreferences())); 
	  	   	
    	   	// Get properties for new reservation. 
    	   	ReservationType modifiedProperties = reservationClient.getReservationProperties(modifiedReservationEpr); 
    	   	ReservationInfo modifiedReservation = new ReservationInfo(); 
    	   	modifiedReservation.setReservation(modifiedProperties); 
    	   	modifiedReservation.setEpr(modifiedReservationEpr); 
    	   	modifiedReservation.setResource(currentResource); 
    	   	
			// Save old reservation for current job. 
			oldReservation = currentJob.getReservation(); 
    	   	
    	   	// Store information about new reservation for current job. 
    	   	currentJob.setReservation(modifiedReservation); 
    	   	
			// Add new reservation to mapping: 
			resource = (GLUECERPType) modifiedReservation.getResource(); 
			reservedJobs = (List) resourceReservations.get(resource); 
			if(reservedJobs == null) {
				reservedJobs = new LinkedList(); 
			}
			reservedJobs.add(currentJob);     						
			resourceReservations.put(resource, reservedJobs); 
			
			// Remove old mapping: 
			resource = (GLUECERPType) oldJob.getReservation().getResource(); 
			reservedJobs = (List) resourceReservations.get(resource); 
			if(reservedJobs == null) {
				// This is a big problem as it should not happen...
				logger.error("Moved a reservation that can not be found in the mappings!"); 
			}
			else {
				reservedJobs.remove(oldJob);
				if(reservedJobs.size() == 0) {
					resourceReservations.put(resource, null); 
				}
			}
			
			// Store old job for next round 
			oldJob = currentJob;
			
		} while(index > 0); 
				
		// No need for an return value, the ContinuingFaultType 
		// contains nextStart for T_augment
	}
    
    /**
     * 
     * @param job 
     * @return Returns information about the created reservation. 
     * @throws ContinuingFaultType If reservation is possible later
     * @throws Exception If reservation never is possible
     */
    public ReservationInfo createReservation(CoallocationInputEntry job, 
    		Calendar early, 
			Calendar late, 
    		ReservationClient reservationClient, 
			GLUECERPType preSelectedResource) throws ContinuingFaultType, 
			Exception {
    	    	
    	// Make it possible to specify which resource to reserve. 
    	List resources = null; 
    	if(preSelectedResource != null) {
    		resources = new LinkedList(); 
    		resources.add(preSelectedResource); 
    	} 
    	else {
    		resources = job.getRankedResources();     	
    	}
    	
    	// Get #CPUs, duration, and flexible from job description
    	int numberOfCPUs = JSDLUtils.getNumberOfCPUs(job.getJobDescription());  	
    	int duration = JSDLUtils.getDuration(job.getJobDescription()); 
    	boolean flexible = JSDLUtils.getFlexible(job.getJobPreferences()); 
    	
    	// Create a reservation starting in the interval [early, late] 
    	// To reduce the amount of reservations created/cancelled, 
    	// this is done using a first-fit algorithm. 
    	
    	// logger.warn("Creating new reservation for job, choosing from " +
    	// 		 resources.size() + " resources"); 
    	
    	boolean done = false; 
     	ReservationInfo reservationInfo = null; 
     	int nextStart = Integer.MAX_VALUE; 
    	for(int i = 0; i < resources.size(); i++) {
    		
    		EndpointReferenceType reservationEPR = null; 
    		GLUECERPType resource = (GLUECERPType) resources.get(i); 
    		
			try { 
				
				// Resolve URL to WSAG factory 
				
				String serviceHost = resource.getCluster(0).getSubCluster(0).getHost(0).getName(); 
				logger.warn("Creating reservation at: " + serviceHost); 
				
				if(serviceHost.equals("toggle02.hpc2n.umu.se") && !this.mayCreateNewAtToggle02) {
					logger.warn("Creation \"failed\" (path augmentation testing)"); 
					throw new Exception ("Creation \"failed\""); 
				}
				
				String factoryURL = ServiceURL.getURL(serviceHost, 
						AgreementFactoryConstants.serviceName).toString(); 
				
				if(!(JobSubmissionUtils.ping(serviceHost))) {
					// logger.debug("Can not find service " + agreementFactoryURI); 
					logger.warn("Resource " + serviceHost + 
					" does not support reservations"); 
					continue; 
				}
				
				// Create reservation
			
				reservationEPR = reservationClient.createReservation(
						factoryURL, 
						numberOfCPUs,
						early, 
						late, 
						duration, 
						flexible);     	   
    	   	} 
    	   	catch (ContinuingFaultType f) {
    	   		// Save best 'next start' of all resources and return it later... 
        		FaultHelper helper = new FaultHelper(f); 
        		String error = helper.getDescriptionAsString();         		
        		logger.warn("Creation failed: " + error);   
        		// Save best 'nextstart' 
        		if(error.startsWith("nextstart ")) {
        			int tmp = Integer.parseInt(error.substring("nextstart ".length())); 
        			if(tmp < nextStart) {
        				nextStart = tmp; 
        			}
        			// logger.warn("Reservation earliest possible at: " + nextStart); 
        		}
        		continue;         		
        	}    	   	    	   	
    	   	catch (Exception e) {
    	   		logger.warn("Creating failed: " + e.getMessage() + 
    	   				" trying next resource");
    	   		continue; 
    	   	}
    	   	
    	   	// Get properties: 
    	   	ReservationType properties = reservationClient.getReservationProperties(reservationEPR); 
    	   	
    	   	// Construct reservation information structure 
    	   	reservationInfo = new ReservationInfo(); 
    	   	reservationInfo.setReservation(properties); 
    	   	reservationInfo.setEpr(reservationEPR); 
    	   	reservationInfo.setResource(resource); 
    	   	
    	   	return reservationInfo; 
    	
    	}
    	
    	if(nextStart == Integer.MAX_VALUE) {
    		logger.warn("Reservation never possible!"); 
    		throw new Exception("Reservation never possible"); 
    	}
    	
    	// return best nextStart 
    	logger.warn("All creations attempts failed..."); 
    	ContinuingFaultType fault = new ContinuingFaultType(); 
    	FaultHelper helper = new FaultHelper(fault); 
    	helper.addDescription("" + nextStart); 
    	throw helper.getFault(); 
    }
       
   
    /**
     * Perform actual coallocation...
     * @param jobs
     * @throws Exception
     */
    public void coallocate(CoallocationInputEntry jobs[], 
    		ReservationClient reservationClient, 
    		Calendar earliestStart, 
			Calendar latestStart, 
    		int tolerance) throws Exception {
    	//, boolean usePathAugmentation) throws Exception { 
    	
    	// Start time window [early, late] 
    	Calendar early = Calendar.getInstance(); 
    	early.setTime(earliestStart.getTime()); 
    	
    	Calendar late = Calendar.getInstance(); 
    	late.setTime(earliestStart.getTime()); 
    	late.add(Calendar.SECOND, tolerance); 
    	
    	// Map resource -> coallocationentries (that have reserved the resource)  
    	Map existingReservations = new HashMap(); 
    	   	
    	boolean done = false; 
    	while(!done) {

    		// nextStart[i] is earliest possible start for job i at any resource
        	int nextStart[] = new int[jobs.length];  
    		    		
    		List jobsToAugment = new LinkedList(); 
     		int nextNewReservation = 0; 
    		int nextAugmentedReservation = 0; 
    		logger.warn("\n ---- Starting new round ---- \n"); 
    		logger.warn("Current window is: [" + early.getTime() 
    				+ " ... " + late.getTime() + "]"); 
    		
    		for(int i = 0; i < jobs.length; i++) {    			    			
    			
    			// Job already has a reservation 
    			if(jobs[i].getReservation() != null) {    				
    				
    				logger.warn("Job " + 
    						jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()
    						+ " already has reservation..."); 
    				// Does reservation start early enough? 
    				Calendar startTime = jobs[i].getReservation().
					getReservation().getExactStart();  
    				if(startTime.after(early) && startTime.before(late)) {
    					logger.warn("Reservation for job " 
    							+ jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName() 
    							+ " starts in time"); 
    					continue; 
    				}    		
    				
    				// MODIFY 
    				
    				// Job has a reservation, but it starts too early
    				// -> Try to move reservation 
    				try { 
    					logger.warn("Trying to move reservation for job " + 
    							jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					EndpointReferenceType newEPR =  reservationClient.modifyReservation(
    							jobs[i].getReservation().getEpr(), 
    							JSDLUtils.getNumberOfCPUs(jobs[i].getJobDescription()), 
								early, 
								late, 
								JSDLUtils.getDuration(jobs[i].getJobDescription()), 
								JSDLUtils.getFlexible(jobs[i].getJobPreferences())); 
    					// Update stored information about reservation
    					ReservationType newReservation = reservationClient.getReservationProperties(newEPR); 
    					jobs[i].getReservation().setEpr(newEPR); 
    					jobs[i].getReservation().setReservation(newReservation); 
    					logger.warn("Reservation successfully moved for job " + 
    							jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName());     
    					logger.warn("Reservation now starts at: " + newReservation.getExactStart().getTime()); 
    					// No need to update mapping here, as reservation stays 
    					// on the same resource...     		
    				} catch (Exception e) {
    					
    					logger.warn("Failed to move reservation" + " for job " + jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					// Failed to move reservation -> reservation cancelled... 
    					// Remove reservation from map
    					GLUECERPType resource = (GLUECERPType) jobs[i].getReservation().getResource(); 
    					List reservedJobs = (List) existingReservations.get(resource);     				
    					if(reservedJobs == null) {
    						// TODO: This should not happen, now what? 
    						logger.error("Removing reservation not found in mapping"); 
    					} 
    					else { 
    						reservedJobs.remove(jobs[i]); 
    						if(reservedJobs.size() == 0) {
    							existingReservations.put(resource, null); 
    						}
    					}
    					// Remove reservation from job. 
    					jobs[i].setReservation(null); 
    				}    				
    				
    				// ENDOF_MODIFY 
    			} 
    			
    			// Job has no reservation (or lost its reservation) 
    			// -> Create new reservation for job. 
    			if(jobs[i].getReservation() == null) { 
    				logger.warn("Job " 
    						+ jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()    						
    						+ " has no reservation..."); 
    				// Try to create new reservation
    				try { 
    					logger.warn("Trying to create new reservation for job " + 
    							jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					this.mayCreateNewAtToggle02 = false; 
    					ReservationInfo newReservation = createReservation(jobs[i], 
    							early, 
								late, 
								reservationClient, 
								null); 
    					this.mayCreateNewAtToggle02 = true; 
    					logger.warn("Successfully created new reservation for job " +
    							jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					
    					// Add reservation to mapping: 
    					GLUECERPType resource = (GLUECERPType) newReservation.getResource(); 
    					List reservedJobs = (List) existingReservations.get(resource); 
    					if(reservedJobs == null) {
    						reservedJobs = new LinkedList(); 
    					}
    					reservedJobs.add(jobs[i]);     						
    					existingReservations.put(resource, reservedJobs); 
    					
    					// Store reservation for job
    					jobs[i].setReservation(newReservation);    				
    					
    				} catch (ContinuingFaultType f) {
    					logger.warn("Could not create new reservation for job " +
    							jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					// Get T_newres_J_i 
    					FaultHelper helper = new FaultHelper(f); 
    					String message = helper.getDescriptionAsString(); 
    					logger.warn("New reservation possible in " + message + " seconds"); 
    					nextNewReservation = Integer.parseInt(message); 
    				}
    				catch (Exception e) {
    					logger.warn("Could not create new reservation for job " +
    							jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					logger.warn("Reservation never possible..."); 
    					nextNewReservation = Integer.MAX_VALUE; 
    				} 
    			} 
    			
    			// Could not create a new reservation
    			// Try augmentation later
    			if(jobs[i].getReservation() == null) {    	
    				jobsToAugment.add(new Integer(i)); 
    			}     			    	 
    			// Test nextstart for augmentation later if augmentation fails
    			// Set nextStart to the value obtained from create(New)Reservation
    			nextStart[i] = nextNewReservation; 
    		
    		} 
    
    		// Perform path augmentation for those jobs lacking a reservation. 
    		if(jobsToAugment.size() > 0) {
    			
    			logger.warn("\nTrying to perform path augmentation"); 
    			
    			logger.warn("Current reservations: "); 
    			for(int i = 0; i < jobs.length; i++) {
    				ReservationInfo info =  jobs[i].getReservation(); 
    				if(info == null) {
    					logger.warn(jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName() +
    							": no reservation"); 
    				}
    				else {
    					logger.warn(jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName() +
    							": " + 
    							((GLUECERPType)info.getResource()).getCluster(0).getSubCluster(0).getHost(0).getName()); 
    				}
    			}
    			
    			logger.warn("Jobs to augment:"); 
    			for(int i = 0; i< jobsToAugment.size(); i++) {
    				logger.warn("Job " + jobs[((Integer)jobsToAugment.get(i)).intValue()].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    			}
    			
    			// Find out whether all jobs lacking reservations 
    			// are connected to the graph where augmented paths can 
    			// be found. In order words: For all jobs, test wheter any other job
    			// has reserved some of the resources this job may use. 
    			// Unless this is the case, path augmentation will never work. 
    			boolean canAugment = true; 
    			for(int i = 0; i < jobsToAugment.size(); i++) {
    				int index = ((Integer) jobsToAugment.get(i)).intValue();     				
    				if(!augmentationPossible(jobs, 
    						jobs[index], 
							existingReservations)) {
    					canAugment = false; 
    					logger.warn("Augmentation not possible for Job " 
    							+ jobs[index].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    				}
    				else {
    					logger.warn("Augmentation may work for Job " 
    							+ jobs[index].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    				}
    			}     	    			    			
    			
    			// All jobs may be possible to augment, now try augmentation. 
    			if(canAugment) {
    				int augmentedNextStartTime; 
    				for(int i = 0; i < jobsToAugment.size(); i++) {
    					
    					int index = ((Integer) jobsToAugment.get(i)).intValue(); 
    					logger.warn("Augmenting path for Job " + 
    							jobs[index].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    					    			
    					try {     			
    						List path = findPath(jobs,
    								jobs[index], 
									early, 
									late, 
									existingReservations); 
    						logger.warn("Augmented path is:"); 
    						printPath(path); 
    						updateReservations(path, 
    								jobs, 
									early, 
									late, 
									reservationClient, 
									existingReservations);     					
    					} catch (ContinuingFaultType f) {
    						logger.warn("Could not augment reservation for job " + 
    							jobs[index].getJobDescription().getJobDescription().getJobIdentification().getJobName()); 
    						// get T_augment_J_i
    						FaultHelper helper = new FaultHelper(f); 
    						String message = helper.getDescriptionAsString(); 
    						logger.warn("Augmented reservation possible " 
    								+ message + " seconds from now"); 
    						
    						// Save next possible start time for path 
    						// augmentation if this is before than the 
    						// next possible time for creating a new reservation
    						augmentedNextStartTime = Integer.parseInt(message);     						
    						if(augmentedNextStartTime < nextStart[index]) {
    							nextStart[index] = augmentedNextStartTime; 
    						}
    						
    						// Don't augment any more reservations as this 
    						// augmentation failed and we will anyhow
    						// not be able to construct a full matching. 
    						break; 
    						
    					} catch (Exception error) {    				    					
    						logger.warn("Path augmentation failed, " +
    						"next possible augmented reservation unknown"); 
    						nextAugmentedReservation = Integer.MAX_VALUE; 					    					
    					}   
    				}
    			} 
    		} 
 
    		boolean allDone = true; 
    		for(int i = 0; i < jobs.length; i++) {
    			if(jobs[i].getReservation() == null) {
    				allDone = false; 
    			}
    		}
    	
    		
    		if(allDone) {
    			done = true;     	    		
    		} 
    		else {
    		    		
    			logger.warn("Moving start time window...");
    			
    			// Find max(T_new_J_i) for all i
    			logger.warn("Earliest start for jobs");  
    			int T_max = 0; 
    			for(int i = 0; i < jobs.length; i++) {
    				logger.warn("Job " 
    						+  jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()
    						+ " " + nextStart[i]); 
    				if(nextStart[i] > T_max) {
    					T_max = nextStart[i]; 
    				}    				
    			}
    			
    			// Move window from [early, late] to 
    			// [late - epsilon, late + T_max]     			
    			late.add(Calendar.SECOND, T_max); 
    			early.setTime(late.getTime()); 
    			early.add(Calendar.SECOND, - tolerance);     			
    			logger.warn("New start time window is:"); 
    			logger.warn("Early: " + early.getTime()); 
    			logger.warn("Late: " + late.getTime());     		    		
    			
    			// Check if window is outside allowed range -> Coallocation failed...     		
    			if(early.after(latestStart)) {
    				logger.warn("Earliest job start is too late, "
    						+ " coallocation attempt failed...");     				
    				
    				logger.warn("Cancelling all reservations"); 
    				for(int i = 0; i < jobs.length; i++) {
    					ReservationInfo reservation = jobs[i].getReservation(); 
    					if(reservation != null) {
    						GLUECERPType glueInfo = (GLUECERPType) jobs[i].getReservation().getResource(); 
    						logger.warn("Cancelling reservation for job "
    								+ jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName() 
									+ " at resource " + glueInfo.getCluster(0).getSubCluster(0).getHost(0).getName()); 
    								
    						reservationClient.destroyReservation(reservation.getEpr());
    					}
    					else { 
    						logger.warn("Job " +
    								jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()
    								+ " has no reservation to cancel..."); 
    					}
    				}
    				
    				ContinuingFaultType f = new ContinuingFaultType(); 
    				FaultHelper helper = new FaultHelper(f); 
    				helper.addDescription("Earliest job start is to late,"
    						+ " coallocation attempt failed..."); 
    				throw helper.getFault(); 
    			}
    			else {
    				logger.warn("Moving window and retrying...");  
    			}
    		}
    	}        	    	
    	
		logger.warn("Final reservations: "); 
		
    	logger.warn("Coallocation attempt successful!"); 
		for(int i = 0; i < jobs.length; i++) {
			ReservationInfo info =  jobs[i].getReservation(); 
			if(info == null) {
				logger.warn(jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName() +
						": no reservation"); 
			}
			else {
				logger.warn(jobs[i].getJobDescription().getJobDescription().getJobIdentification().getJobName() +
						": " + 
						((GLUECERPType)info.getResource()).getCluster(0).getSubCluster(0).getHost(0).getName()); 
			}
		}
    	
    }
    
    /**
     * Performs resource discovery and resource ranking for each job
     * description. Next creates a set of reservations that fulfills
     * the requirements of the jobs. 
     * Finally submits all jobs to their selected (reserved) resources 
     * and confirms all reservations. 
     * 
     * @param jobDescriptions
     * @param indexServerURIs
     * @param jobPreferences
     * @param clientURL
     * @param clientStagesInput
     * @return
     * @throws Exception
     */
    public SubmittedJob[] coallocateJob(JobDefinition_Type jobDescriptions[], 
            String indexServerURIs[], 
            JobPreferencesType jobPreferences[], 
            String clientURL, 
			boolean clientStagesInput, 
			int tolerance) throws Exception {
    	
    	try {
 
    		Broker broker = new Broker(this.numberOfThreads); 
    		ReservationUtils reservationUtils = new ReservationUtils(); 
    		broker.setReservationUtils(reservationUtils); 
    		ReservationClient reservationClient = new ReservationClient(); 
    		
    		// Verify that all job description contains all required 
    		// attributes (currently, only testing existence of "Executable").
    		// This call throws InvalidJobDescriptionFaultType upon error.
    		for(int i = 0; i < jobDescriptions.length; i++) {
    			broker.verify(jobDescriptions[i]); 
    		}    		    		 
    		
    		// Fix job preferences, that is set an allowed start time delay 
    		// for those jobs that does not specify one. For such jobs, 
    		// the allowed start time delay is the max of that of the other 
    		// jobs, of a default value ([0, 1 hour]) 
    		JobPreferencesType newPreferences[] = fixJobPreferences(jobPreferences, 
    				jobDescriptions.length); 
    
       		// For each job, filter out the resources not apt for the job, 
    		// and rank the remaining ones.     		
    		List rankedResources[] = rankResources(jobDescriptions, 
    				indexServerURIs, 
					newPreferences, 
					broker); 
    	
    		CoallocationInputEntry input[] = new CoallocationInputEntry[jobDescriptions.length]; 
    		for(int i = 0; i < jobDescriptions.length; i++) {
    			input[i] = new CoallocationInputEntry(); 
    			input[i].setJobDescription(jobDescriptions[i]); 
    			input[i].setRankedResources(rankedResources[i]); 
    			input[i].setJobPreferences(newPreferences[i]); 
    			input[i].setReservation(null); 
    		}
    		
    		logger.warn("Job start delay preferences"); 
    		for(int i = 0; i < input.length; i++) {
    			try {
    				JobPreferencesType jobPrefs= input[i].getJobPreferences(); 
    				JobRequirementType reqs = input[i].getJobPreferences().getJobRequirement(); 
    				logger.warn(reqs.getEarliestAbsoluteStart().getTime()
    						+ " ... " 
    						+ reqs.getLatestAbsoluteStart().getTime()); 
    			} catch (Exception e) {
    				logger.error("Ooops: " + e.getMessage()); 
    			}
    		}

    		/* 
    	     * Reorder entries to ensure that jobs requesting more powerful 
    	     * resources are reserved first. We have three sorting keys, listed
		     * below in order of importance: 
		     * 1) Number of CPUs 
		     * 2) Runtime
		     * 3) Amount of memory 
		     */ 
    		
    		// Sort with least significant key first. As Arrays.sort 
    		// is stable, the tripple key-sorting is achieved. 
    		Arrays.sort(input, new MemoryComparator());            		        	        
    		Arrays.sort(input, new RunTimeComparator());            		        	        
    		Arrays.sort(input, new NumberOfCPUsComparator());

    		// Testing only, tag jobs with their index. 
    		// N.B: It is rather important to do this _after_ the above sortings :-)
    		for(int i = 0; i < input.length; i++) {
    			JobIdentification_Type jobId = new JobIdentification_Type(); 
    			jobId.setJobName("<Job " + i + ">"); 
    			input[i].getJobDescription().getJobDescription().setJobIdentification(jobId); 
    		}
    		
    		logger.warn("After rearrangement, we have the following job order"); 
    		logger.warn("JobIndex #CPUs runTime #memory"); 
    		for(int i = 0; i < input.length; i++) {
    			logger.warn("" + 
    					input[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()
    					+ " " 
    					+ JSDLUtils.getNumberOfCPUs(input[i].getJobDescription()) + " "
    					+ JSDLUtils.getDuration(input[i].getJobDescription()) + " "
						+ JSDLUtils.getMemory(input[i].getJobDescription())); 				
    		}
    		
    		// Output ranked resources
    		for(int i = 0; i < input.length; i++) {
     			logger.warn("The preferred resources for job " 
     					+ input[i].getJobDescription().getJobDescription().getJobIdentification().getJobName());      		    		
    			List resources = input[i].getRankedResources(); 
    			for(int j = 0; j < resources.size(); j++) {
    				GLUECERPType resource = (GLUECERPType) resources.get(j);
    				logger.warn(resource.getCluster(0).getSubCluster(0).getHost(0).getName()); 
    			}			
    		}
        		
    		// Determine earliest and latest allowed start times. 
    		// TODO: For now, just take smallest possible window. 
    		// For later, allow individual offset from window... 
    		Calendar earliestStart = null; 
    		Calendar latestStart = null; 
    		
    		for(int i = 0; i < input.length; i++) {
    			JobRequirementType requirements = input[i].getJobPreferences().
				getJobRequirement(); 
    			
    			Calendar early = requirements.getEarliestAbsoluteStart(); 
    			if(earliestStart == null || early.after(earliestStart)) {
    				earliestStart = early; 
    			}
    			Calendar late = requirements.getLatestAbsoluteStart(); 
    			if(latestStart == null || late.before(latestStart)) {
    				latestStart = late; 
    			}
    		}
    		
    		// Perform coallocation (coreservation) 
    		try { 
        		
        		logger.warn("For coallocation, we have the following window:"); 
        		logger.warn("Early: " + earliestStart.getTime() 
        				+ " Late: " + latestStart.getTime()); 
        		
        		coallocate(input, 
        				reservationClient, 
						earliestStart, 
						latestStart, 
						tolerance); // ,  usePathAugmentation); 
        		/* 
        		logger.warn("We have the following reservations:"); 
        		for(int i = 0; i < input.length; i++) {
        			logger.warn(input[i].getReservation().getReservation().getLocalID()); 
        		}*/         		        		        	        		
        	
        	} catch (ContinuingFaultType f) {
        		logger.warn("Coallocation attempt failed"); 
        		throw f; 
        	} catch (Exception e) {
        		logger.warn("Oops, reservations failed:"); 
        		throw e; 
        	}
        	
        	logger.warn("All subjobs successfully coallocated:\n"); 
        	for(int i = 0; i < input.length; i++) {
        		logger.warn("Start time for job "
        				+ input[i].getJobDescription().getJobDescription().getJobIdentification().getJobName()
						+ " : "
        				+ input[i].getReservation().getReservation().getExactStart().getTime());
        	}
        	
        	// Dispatch all subjobs
            Dispatcher dispatchers[] = SubmitterUtils.getDispatchers(this.dispatcherClasses); 
            SubmittedJob submittedJobs[] = new SubmittedJob[input.length]; 

            for(int i = 0; i < input.length; i++) {
            	
            	Object jobDescription = input[i].getJobDescription(); 
            	ReservationInfo reservation = input[i].getReservation(); 
            	Object reservedResource = reservation.getResource(); 
            	
            	// TODO: Implement server-side file staging? 
            	// Coallocation client can stage local input files, 
            	// so in pracise, this may not be needed. 
            	boolean dispatched = false; 
            	for(int j = 0; j < dispatchers.length && !dispatched; j++) {
            		logger.info("Dispatching job using: " + 
            				dispatchers[j].getClass()); 
            		try {            			
            			Object jobId = dispatchers[j].requestJob(reservedResource, 
            					jobDescription, 
								reservation,
								clientURL, 
								clientStagesInput); 
            			
            			logger.warn("Job dispatched with id: " + jobId); 
            			submittedJobs[i] = new SubmittedJob(); 
            			submittedJobs[i].setLocalID(jobId.toString()); 
            			submittedJobs[i].setJobDescription((JobDefinition_Type)jobDescription); 
            			submittedJobs[i].setReservationEPR(reservation.getEpr()); 
            			
            			dispatched = true; 
            			
            		} catch (Exception e) {        			
            			logger.error("Could not dispatch job using: " 
            					+ dispatchers[j].getClass()); 
            		}
            	} 
            	// A subjob submission failed 
            	if(!dispatched) {
            		logger.error("Could not dispatch job " 
            				+ input[i].getJobDescription().getJobDescription().getJobIdentification().getJobName());
            		logger.error("Cancelling previously submitted jobs"); 
         
            		// Cancel all already submitted jobs...             	
            		for(int k = 0; k < i; k++) {
            			logger.error("Cancelling job: " + submittedJobs[k].getLocalID()); 
            			ClientUtils.cancelJob(submittedJobs[k].getLocalID(), 
            					this.stagingTargetResolverClasses); 
            		}
            		
               		throw new Exception("Could not dispatch job " + i); 
            	}
            } 

        	logger.warn("All subjobs successfully dispatched"); 
        	logger.warn("Confirming reservations");         	
        	for(int i = 0; i < submittedJobs.length; i++) {
        		reservationClient.commitReservation(submittedJobs[i].getReservationEPR()); 
        	}
        	logger.warn("Done confirming reservations"); 
        
            return submittedJobs; 
    		
    		
    	} catch (Exception e) {
    		logger.error("Coallocation attempt failed:");
    		logger.error(e.getMessage()); 
    		throw e; 
    	}            	
    }
}


